#ifndef __DATABUFFERPAR_H__
#define __DATABUFFERPAR_H__

#include "../std.h"
#include <pthread.h>
#include "dataspeed.h"

class CheckSum;

/*
  Buffer to use during data transfer.
*/

class DataBufferPar {
 private:
  /* used to check if configuration cahnged */
  int set_counter;
  /* general purpose mutex and condition used to 
     achieve thread safety */
  pthread_mutex_t lock;
  pthread_cond_t  cond;
  /* structure do describe status of every buffer */
  typedef struct {
    /* buffer address in memory */
    char* start;
    /* true if taken by application for filling */
    bool taken_for_read;
    /* true if taken by application for emptying */
    bool taken_for_write; 
    /* size of buffer */
    unsigned int size;
    /* amount of information stored */
    unsigned int used;
    /* offset in file or similar, has meaning only for application */
    unsigned long long int offset;
  } buf_desc;
  /* amount of data passed through buffer (including current stored). 
    computed using offset and size. gaps are ignored. */
  unsigned long long int eof_pos;
  /* list of controlled buffers */
  buf_desc* bufs;
  /* amount of controlled buffers */
  int bufs_n;
  /* set to true if application's reading(filling) part 
     won't use buffer anymore */
  bool eof_read_flag;
  /* same for writing(emptying) part */
  bool eof_write_flag;
  /* reading part of application experienced error */
  bool error_read_flag;
  /* same for writing part */
  bool error_write_flag;
  /* error was originated in DataBufferPar itself */
  bool error_transfer_flag;
  /* wait for any change of buffers' status */
  bool cond_wait(void);
  /* pointer to object to compute checksum */
  CheckSum *checksum;
  unsigned long long int checksum_offset;
  bool checksum_ready;
  
 public:
  /* object to control transfer speed */
  DataSpeed speed;
  operator bool(void) { return (bufs!=0); };
  /* 
    Contructors and destructor.
    size - size of every buffer
    blocks - number of buffers
    cksum - object to compute checksum. Should not be destroyed
      till DataBufferPar itself.
  */
  DataBufferPar(unsigned int size = 65536,int blocks = 3);
  DataBufferPar(CheckSum *cksum,unsigned int size = 65536,int blocks = 3);
  ~DataBufferPar(void);
  /*
    Reinitialize buffer with different parameters.
  */
  bool set(CheckSum *cksum = NULL,unsigned int size = 65536,int blocks = 3);
  char* operator[](int n);
  /* provide buffer for READING INTO IT 
       handle - number of buffer. To be used to reffer to buufer.
       length - size of buffer
       wait - if true and there are no free buffers, method will
         wait for one.
     returns:
       true - success
       false - failure
  */
  bool for_read(int &handle,unsigned int &length,bool wait);
  /* 
    Inform object data was read into buffer.
      handle - buffer's number
      buf - address of buffer
      length - amount of data
      offset - offset in stream, file, etc.
  */
  bool is_read(int handle,unsigned int length,unsigned long long int offset);
  bool is_read(char* buf,unsigned int length,unsigned long long int offset);
  /* provide buffer for WRITING FROM IT 
        accepted atguments and result are same as for for_read */
  bool for_write(int &handle,unsigned int &length,unsigned long long int &offset,bool wait);
  /* 
    Inform object data was written from buffer.
  */
  bool is_written(int handle);
  bool is_written(char* buf);
  bool is_notwritten(int handle);
  bool is_notwritten(char* buf);
  /*
    Inform object of read/write part of aplication is not
    going to use buffer anymore.
  */
  void eof_read(bool);
  void eof_write(bool);
  /* returns if error was detected */
  /*
    Inform object of error accured in read/write part of aplication 
  */
  void error_read(bool);
  void error_write(bool);
  /* return true if object was informed about coresponding change/problem */
  bool eof_read(void);
  bool eof_write(void);
  bool error_read(void);
  bool error_write(void);
  bool error_transfer(void);
  bool error(void);
  /* wrapper for cond_wait with locking */
  bool wait(void);
  /* wait till there are no more used buffers left in object */
  bool wait_used(void);
  /* returns true if there were no problem with checksum computing */
  bool checksum_valid(void);
  const CheckSum* checksum_object(void);
  bool wait_eof_read(void);
  bool wait_eof_write(void);
  bool wait_eof(void);
  unsigned long long int eof_position(void) const { return eof_pos; };
  unsigned int buffer_size(void);
};

#endif

